;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "org-appear" "20260716.2120"
  "Auto-toggle Org elements."
  '((emacs "29.1")
    (org   "9.3"))
  :url "https://github.com/awth13/org-appear"
  :commit "77d23efec5f5c25fc0798364d2b51a3ce3d8d518"
  :revdesc "77d23efec5f5"
  :authors '(("Alice Istleyeva" . "awth13@gmail.com"))
  :maintainers '(("Alice Istleyeva" . "awth13@gmail.com")))
